using System;
using System.Collections.Generic;
using System.Threading;
using NUnit.Framework;
using NUnit.Framework.SyntaxHelpers;

namespace DelegateSubstitutionVSIfStatements
{
	class Program
	{
		static void Main(string[] args)
		{
			Average(new Using(10000).Delegates, 10000, 100, 10000);
			Average(new Using(10000).Ifs, 10000, 100, 10000);

			Average(new Using(5000).Delegates, 10000, 100, 5000);
			Average(new Using(5000).Ifs, 10000, 100, 5000);
			
			Average(new Using(2).Delegates, 10000, 100, 2);
			Average(new Using(2).Ifs, 10000, 100, 2);

			Average(new Using(1).Delegates, 10000, 100, 1);
			Average(new Using(1).Ifs, 10000, 100, 1);
		}

		private static void Average(ThreadStart start, int calls, int times, int initialCalls)
		{
			List<double> runTimes = new List<double>();
			for(int i = 0; i<times;i++)
				runTimes.Add(Run(start, calls));

			Console.WriteLine("{3}:{2} calls using {1} averaging {0} seconds.",
				AverageTime(runTimes), start.Method.Name.PadRight(9), calls, initialCalls);
		}

		private static double AverageTime(ICollection<double> runTimes)
		{
			double sum = 0;
			foreach (double value in runTimes)
				sum += value;
			return sum/runTimes.Count;
		}

		private static double Run(ThreadStart start, int times)
		{
			DateTime started = DateTime.Now;
			for(int i = 0; i< times; i++)
				start();
			return DateTime.Now.Subtract(started).TotalSeconds;
		}
	}

	class Using
	{
		private delegate void ExecuteSomething();
		private ExecuteSomething _executeMe;
		private readonly int _runTaskOneTimes;
		private int _taskOneRuns = 0;
		private int _taskTwoRuns = 0;

		public Using(int initialItems)
		{
			_executeMe = InitialDelegate;
			_runTaskOneTimes = initialItems;
		}

		public int TaskOneRuns
		{
			get { return _taskOneRuns; }
		}

		public int TaskTwoRuns
		{
			get { return _taskTwoRuns; }
		}

		public void Ifs()
		{
			if (!TaskOneIsComplete())
				TaskOne();
			TaskTwo();
		}

		public void Delegates()
		{
			_executeMe();
		}
		
		private bool TaskOneIsComplete()
		{
			return _taskOneRuns >= _runTaskOneTimes;
		}

		private void InitialDelegate()
		{
			TaskOne();
			TaskTwo();

			if (TaskOneIsComplete())
				_executeMe = AfterDelegate;
		}

		private void AfterDelegate()
		{
			TaskTwo();
		}

		private void TaskOne()
		{
			Math.Sqrt(DateTime.Now.Ticks);
			_taskOneRuns++;
		}

		private void TaskTwo()
		{
			Math.Sqrt(DateTime.Now.Ticks);
			_taskTwoRuns++;
		}
	}

	[TestFixture]
	public class UsingTest
	{
		[Test]
		public void TestDelegates()
		{
			Using u = new Using(1);
			u.Delegates();
			u.Delegates();
			Assert.That(u.TaskOneRuns, Is.EqualTo(1));
			Assert.That(u.TaskTwoRuns, Is.EqualTo(2));
		}

		[Test]
		public void TestIfs()
		{
			Using u = new Using(1);
			u.Ifs();
			u.Ifs();
			Assert.That(u.TaskOneRuns, Is.EqualTo(1));
			Assert.That(u.TaskTwoRuns, Is.EqualTo(2));
		}
	}
}
